"""Door module"""
from game.model.element import Element


class Door(Element):
    """Class to initialize any kind of door, e.g. a trapdoor, with its own name and description"""
    def __init__(self, name, description, **kwargs):
        self.name: str = name
        self.description: str = description
        self.open: bool = kwargs.get("open", False)
        self.key: str = kwargs.get("key", None)
        self.lockable: bool = kwargs.get("lockable", self.key is not None)
        self.locked: bool = kwargs.get("locked", True if self.key else False)
        """Value must be the name of the key as a string"""
        self.connects: list = kwargs.get("connects", [])
        """Values in list must be names of locations as strings"""
        super().__init__(name, description, **kwargs)



"""Element module"""


class Element:
    """Class to initialize any element with its own name and description"""
    def __init__(self, name, description, **kwargs):
        self.name: str = name
        self.description: str = description
        self.visible: bool = kwargs.get("visible", True)
        self.class_name: str = kwargs.get("class_name", type(self).__name__)
        self.preposition: str = kwargs.get("preposition", "in")
        self.sound: str = kwargs.get("sound", None)
        self.smell: str = kwargs.get("smell", None)
        self.contents: list = kwargs.get("contents", [])
        """Values in list must be Element objects"""



"""helper methods module"""
def isinstanceorsubclass(__obj: object, __class: type):
    """Check if objects is instance or subclass of type"""
    return isinstance(__obj, __class) or issubclass(__obj.__class__, __class)



"""test dungeon master module"""
import unittest
from unittest.mock import Mock
from parameterized import parameterized
from game.controller.dungeon_controller import DungeonMaster

class TestDungeonMaster(unittest.TestCase):
    """Test DungeonMaster class"""

    def setUp(self):
        """Set up environment required for tests"""
        self.dungeon_master = DungeonMaster()

    @parameterized.expand([
        ["You enter a quirky test location.", "You enter a quirky test location.\n\n"],
        ["", ""],
    ])
    def test_brief(self, brief, expected):
        """Test brief method"""
        mock_location = Mock()
        attrs = {
            "brief": brief,
            "visited": False
        }
        mock_location.configure_mock(**attrs)
        self.dungeon_master.all_name_locations.append(("test", mock_location))
        self.assertEqual(self.dungeon_master.brief("test"), expected)



"""Location module"""
from game.model.element import Element


class Location(Element):
    """Class to initialize any kind of place, e.g. a room, with its own name and description"""
    def __init__(self, name, description, **kwargs):
        self.name: str = name
        self.description: str = description
        self.visited: bool = kwargs.get("visited", False)
        self.exits: dict = kwargs.get("exits", {})
        """Keys must be directions e.g. west, values must be names of locations e.g. cellar"""
        self.has_light: bool = kwargs.get("has_light", False)
        self.brief: str = kwargs.get("brief", None) # ToDo: implement
        self.needs_rope: bool = kwargs.get("needs_rope", False)
        Element.__init__(self, name, description, **kwargs)



"""Tool module"""
from game.model.element import Element
from game.model.thing import Thing


class Tool(Thing, Element):
    """Class to initialize any kind of tool or weapon, e.g. a sword,
    with its own name and description"""
    def __init__(self, name, description, **kwargs):
        self.damage: int = kwargs.get("damage", 10)
        self.durability: int = kwargs.get("durability", 100)
        Thing.__init__(self, name, description, **kwargs)
        Element.__init__(self, name, description, **kwargs)



"""main module"""
from game.view.console_view import ConsoleView
from game.controller.dungeon_controller import DungeonMaster
import pkg_resources


def run():
    """main method that starts the program"""
    dungeon_master = DungeonMaster()
    scenario_filename = pkg_resources.resource_filename("game.data", "scenario.json")
    dungeon_master.load(scenario_filename)
    view = ConsoleView(dungeon_master)
    view.start_view()



"""Activator module"""
from game.model.enums.activator_type import ActivatorType
from game.model.element import Element


class Activator(Element):
    """Class to initialize any kind of activator, e.g. a button, with its own name and description"""
    def __init__(self, name, description, **kwargs):
        self.name: str = name
        self.description: str = description
        self.type: str = kwargs.get("type", ActivatorType.PRESS)
        self.is_on: bool = kwargs.get("is_on", False)
        self.turn_on_method_name: str = kwargs.get("turn_on_method_name", None)
        self.turn_off_method_name: str = kwargs.get("turn_off_method_name", None)
        super().__init__(name, description, **kwargs)



"""test activator handler module"""
import unittest

from game.controller.activator_controller import ActivatorHandler
from game.data.scenario_texts import BUTTON_GOES_DARK, BUTTON_LIGHTS_UP

class TestActivatorHandler(unittest.TestCase):
    """Test ActivatorHandler class"""

    def setUp(self):
        """Set up environment required for tests"""
        self.activator_handler = ActivatorHandler()

    def test_bedroom_button_on(self):
        """Test bedroom button on method"""
        self.assertEqual(self.activator_handler.bedroom_button_on(), BUTTON_LIGHTS_UP)

    def test_bedroom_button_off(self):
        """Test bedroom button off method"""
        self.assertEqual(self.activator_handler.bedroom_button_off(), BUTTON_GOES_DARK)



"""texts module"""
# introduction
INTRODUCTION = "\nWelcome to Hemonaluto! A curious magical world with a dark and powerful secret."
PLEASE_TYPE = "Type any command to start exploring your new environment."

# quit message
QUIT_MESSAGE = "They will miss you."

# indicator for the user to recognize it's their turn to type something
INPUT_INDICATOR = "\n>"

# examine command
EXAMINE_COMMAND_NAME = "examine"
EXAMINE_COMMAND_DESCRIPTION = "Examine anything in the game"
EXAMINE_COMMAND_HELP = "Type examine followed by the thing you want to examine."\
        "\nExamples:\nexamine door\nexamine location\nexamine knife"

# examine location
LOCATION_PREFIX = "You are in "
LOCATION_SUFFIX = "\nYou look around you and you see:"

# directions
NORTH = "north"
EAST = "east"
SOUTH = "south"
WEST = "west"
NORTHEAST = "northeast"
NORTHWEST = "northwest"
SOUTHEAST = "southeast"
SOUTHWEST = "southwest"
UP = "up"
DOWN = "down"
INVALID_DIRECTION = "You can't go there."

# door
LOCKED_DOOR = "There is a locked door blocking the path."
def door_unlocked(door_name):
    """Returns the message for when the player unlocks a specific door."""
    return f"You have unlocked the {door_name}."
def door_not_locked(door_name):
    """Returns the message for when the player tries to unlock a\
    specific door that already is unlocked."""
    return f"{door_name.capitalize()} is already unlocked."
KEY_MISSING = "You don't have the required key for that."

# thing in container
def element_in_container(things_as_string, preposition, container):
    """Returns a description for a thing within a container"""
    return f"    There is a {things_as_string} {preposition} the {container}."

# element not found
def element_not_found(element_name):
    """Returns element not found message"""
    return f"You couldn't find the {element_name}."

# element not in inventory
def element_not_in_inventory(element_name):
    """Returns element not in inventory message"""
    return f"You don't have the {element_name} in your inventory."

# picked up element
def picked_up_element(element_name):
    """Returns the information that the player picked up a specific element"""
    return f"You picked up the {element_name}."

# generic location
GENERIC_LOCATAION_NAME = "location"

# greetings
GREETINGS = ["Hi.", "Hello.", "Greetings"]

# swear response
SWEAR_RESPONSE = "No. More. Saying. Cuss words! It. Is. Not. Good. I\'m putting a video on YouTube \
about no more saying cuss words. No more saying cuss words guys! It's inappropriate and violent! \
If you say a cuss word then you're like, going to jail, and you're like, and when you go to jail, \
i- ba- when you go to jail, if you say, if you say a cuss word you go to jail and if you go to \
jail cause you said a cuss word, then... You're only gonna eat BROCCOLI and OTHER VEGETABLES \
for your WHOLE LIFE. You don't want to eat vegetables. Sometimes people like eating sweets but, \
I eat broccoli. So, I'm okay with broccoli but I do not want to go to jail. You can not go to jail. \
And saying cuss words is ILLEGAL. They are now gonna make a law about that. It is illegal, it is \
inappropriate, it is really violent. I better warn my school about that."

# jump response
JUMP_RESPONSE = "You jump up and down."

# saved game message
SAVED_GAME_MESSAGE = "Saved the game!"

# loaded save message
LOADED_SAVE_MESSAGE = "Loaded last save!"

# failed save message
FAILED_SAVE_MESSAGE = "Failed to load last save."

# scenario loaded message
SCENARIO_LOADED = "Loaded scenario."

# nothing response
NOTHING_RESPONSES = ["What do sea monsters eat for dinner?\n\nFish and ships.",\
"Why did the necromancer kill his depressed allies?\n\nHe wanted to raise their spirits.",\
"Why do dwarven bards sound better by candlelight?\n\nYou can shove the wax in your ears."]

# hit target response
def hit_target(target_name):
    """Returns hit target message"""
    return f"Your weapon hit the {target_name}."

# threw thing at nothing specific response
THREW_AT_NOTHING = "You threw it on the ground."

# closed door or chest response
CLOSED = "Closed it."

# tries to close something unopenable response
NOT_OPENABLE = "You can't close that."

# tries to read something unreadable response
NOT_READABLE = "You can't read that."

# tries to throw something unthrowable
NOT_THROWABLE = "You can't throw that."

# tries to turn on something that is already on
ALREADY_ON = "That's already turned on."

# tries to turn off something that is already off
ALREADY_OFF = "That's already turned off."

# turned on an activator
TURNED_ON = "Turned on."

# turned off an activator
TURNED_OFF = "Turned off."

# tries to do an impossible action
ACTION_NOT_POSSIBLE = "You can't do that."

# action failed
ACTION_FAILED = "Action failed."

# tries activator that does nothing
NOTHING_HAPPENS = "Nothing happens."

# reveal element
def reveal_element(moved_item, revealed_item):
    """Revealing element message"""
    return f"You move the {moved_item} and reveal {revealed_item}"

# attacks without specifying weapon
WEAPON_NOT_SPECIFIED = "You didn't specify a weapon."

# eating food message
def eat_food(food_name, taste):
    """Eating food message"""
    return f"You eat the {food_name}. It tastes {taste}."

# shouting response
SHOUT_RESPONSE = "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"

# tied element to target
def tie_rope_to_target(target):
    """Tying rope to something message"""
    return f"You tie the rope to the {target}."

# that won't hold
THAT_WONT_HOLD = "That won't hold your weight."

# target not specified
TARGET_NOT_SPECIFIED = "Target not specified."

# untie
UNTIE = "You untie it."

# climbing down text
CLIMBING_DOWN = "You climb down the rope.\n"

# tries to go down when there is no tied rope in the room
NO_TIED_ROPE = "There is a steep way down but you'll need something to lower yourself."

# tries to tie rope to something you can't tie ropes to
CANT_TIE_TO_ELEMENT = "You can't tie the rope to that."

# tries to untie an already untied rope
ALREADY_UNTIED = "No need for that, it's already untied."

# tries to break a thing that has no break method
CANT_BREAK = "Breaking that would be pointless."

# tries to attack something without a valid tool
NEEDS_TO_BE_TOOL = "Try using a tool or weapon instead."

# door leads to message
def door_leads_to(directions):
    """describes where a door leads to"""
    description = ""
    for direction in directions:
        description += f"\n    The door leads to the {direction}."
    return description

# hears nothing message
SILENCE = "You hear nothing specific."

# smells nothing message
NO_SMELLS = "You smell nothing specific."

# noises description
def noises_description(noises):
    """Describes noises"""
    description = "You hear:"
    for noise in noises:
        description += f"\n    {noise}"
    return description

# smell description
def smell_description(smells):
    """Describes smells"""
    description = "You smell:"
    for smell in smells:
        description += f"\n    {smell}"
    return description

# hiding
def entering_thing(thing):
    """Entering thing message"""
    return f"Hiding in {thing}"
NOT_ENTERABLE = "You can't enter that."
SPECIFIY_HIDING_PLACE = "Hidingplace not specified."
CANT_SEE_LOCATION_FROM_HIDING = "You can't see anything from your hididng spot."
APPEARING = "You leave your hiding place."
NOT_HIDING = "You're not hiding in the first place."

# done generic action
DONE = "Done."



"""scenario texts module"""
# player
PLAYER_NAME = "Player"
PLAYER_DESCRIPTION = "A slim person with fiery blue eyes and pitch-black hair."

#########################################################################

# cellar
CELLAR_NAME = "Cellar"
CELLAR_DESCRIPTION = "A dusty abandoned cellar with an unusable broken staircase leading up.\n"
CELLAR_BRIEF = "You find yourself in a repurposed cellar. There are dustparticals dancing rhythmically around you; there is magic in the air."

# cellar altar
CELLAR_ALTAR_NAME = "altar"
CELLAR_ALTAR_DESCRIPTION = "A large round object covered in a tarp."

# cellar skeleton
CELLAR_SKELETON_NAME = "skeleton"
CELLAR_SKELETON_DESCRIPTION = "A skeleton."

#########################################################################

# dining room
DINING_ROOM_NAME = "Dining Room"
DINING_ROOM_DESCRIPTION = "A large L shaped room with two closed windows illuminating a homey interior."
DINING_ROOM_BRIEF = "As you enter the room you get greeted by the warmth of a crackling fireplace in the corner and the smell of a fresh breakfast."

# rope
ROPE_NAME = "rope"
ROPE_DESCRIPTION = "A long rope."

# dining room trapdoor
DINING_ROOM_TRAPDOOR_NAME = "trapdoor"
DINING_ROOM_TRAPDOOR_DESCRIPTION = "A wooden trapdoor."

# dining room food
DINING_ROOM_FOOD_NAME = "breakfast"
DINING_ROOM_FOOD_DESCRIPTION = "A hearty breakfasat."
DINING_ROOM_FOOD_TASTE = "amazing"
DINING_ROOM_FOOD_SMELL = "A pleasent smell resembling eggs and bacon."

# dining room table
DINING_ROOM_TABLE_NAME = "table"
DINING_ROOM_TABLE_DESCRIPTION = "A round wooden table."

# dining room plate
DINING_ROOM_PLATE_NAME = "plate"
DINING_ROOM_PLATE_DESCRIPTION = "A ceramic plate."

# dining room fireplace
DINING_ROOM_FIREPLACE_NAME = "fireplace"
DINING_ROOM_FIREPLACE_DESCRIPTION = "A lit fireplace."
DINING_ROOM_FIREPLACE_SOUND = "The crackle of a fire."

# breakfast knife
BREAKFAST_KNIFE_NAME = "knife"
BREAKFAST_KNIFE_DESCRIPTION = "A small blunt knife."

# dining room crate
DINING_ROOM_CRATE_NAME = "crate"
DINING_ROOM_CRATE_DESCRIPTION = "A heavy crate filled to the brim with dirt."

# plate
PLATE_BREAKS = "The plate shatters into a dozen pieces."

#########################################################################

# bedroom
BEDROOM_NAME = "Bedroom"
BEDROOM_DESCRIPTION = "A room designed for humans to rest."
BEDROOM_BRIEF = "You are standig in a beautiful forest clearing with three beings infront of you. Their form is fuzzy and blurred and you can't make out their true appearence. They want to tell you something important but as they are about to do so you hear a distant clinking followed by a door knocking. You realize you're in a dream and you wake up covered in warm cozy bedsheets."

# bedroom key text
BEDROOM_KEY_TEXT = "B"

# pile of dust
BEDROOM_PILE_OF_DUST_NAME = "pile of dust"
BEDROOM_PILE_OF_DUST_DESCRIPTION = "A large pile of dust."

# bedroom button
BEDROOM_BUTTON_NAME = "button"
BEDROOM_BUTTON_DESCRIPTION = "A big decorated wooden button attached to the wall next to the bed."

# bed
BED_NAME = "bed"
BED_DESCRIPTION = "A comfy wooden red bed."

# bedroom door
BEDROOM_DOOR_NAME = "door"
BEDROOM_DOOR_DESCRIPTION = "A mundane wooden door."

# bedroom key
BEDROOM_KEY_NAME = "crude key"
BEDROOM_KEY_DESCRIPTION = "A key with a cursive letter B inscribed on it."

# bedroom hook
BEDROOM_HOOK_NAME = "hook"
BEDROOM_HOOK_DESCRIPTION = "A blunt wooden hook attached to the wall."

# bedroom rug
BEDROOM_RUG_NAME = "rug"
BEDROOM_RUG_DESCRIPTION = "A large soft red rug."

# bedroom button
BUTTON_LIGHTS_UP = "The button lights up."
BUTTON_GOES_DARK = "The button goes dark."



"""save handler module"""
import json
from typing import Dict, List, Tuple
from game.helper_methods import isinstanceorsubclass
from game.model.location import Location
from game.model.player import Player
from game.model.rope import Rope
from game.model.thing import Thing
from game.model.tool import Tool
from game.model.activator import Activator
from game.model.animate import Animate
from game.model.chest import Chest
from game.model.door import Door
from game.model.food import Food
from game.model.element import Element


class SaveHandler():
    """Class to save and load the game state"""
   
    class ElementEncoder(json.JSONEncoder):
        """json encoder for elements"""
        def default(self, o):
            return o.__dict__

    def save(self, all_name_locations: List[Tuple[str, Location]], filename):
        """Saves the game"""
        locations = []
        for location in all_name_locations:
            locations.append(location[1])
        with open(filename, "w", encoding="UTF-8") as savefile:
            json.dump(locations, savefile, indent=4, cls=self.ElementEncoder)
        return True

    def load(self, filename: str):
        """Loads the save file"""
        all_name_locations = []
        player_location = None
        with open(filename, "r", encoding="UTF-8") as savefile:
            location_dictionaries = json.load(savefile)
            for location_dictionary in location_dictionaries:
                contents_dictionary = location_dictionary["contents"]
                location = Location(**location_dictionary)
                location.contents = self.dictionary_to_elements(contents_dictionary)
                all_name_locations.append((location.name, location))
                for element in location.contents:
                    if isinstanceorsubclass(element, Player):
                        player_location = location
            return (all_name_locations, player_location)

    def dictionary_to_elements(self, contents_dictionary_list: Dict):
        """Converts dictionary to elements"""
        converted_contents = []
        for element_dictionary in contents_dictionary_list:
            class_name = element_dictionary["class_name"]
            element = eval(class_name)(**element_dictionary)
            if len(element_dictionary["contents"]) > 0:
                element.contents = self.dictionary_to_elements(element_dictionary["contents"])
            converted_contents.append(element)
        return converted_contents



"""Thing module"""
from game.model.element import Element


class Thing(Element):
    """Class to initialize any kind of object, e.g. a bottle, with its own name and description"""
    def __init__(self, name, description, **kwargs):
        self.name: str = name
        self.description: str = description
        self.fixed: bool = kwargs.get("fixed", False)
        self.moved: bool = kwargs.get("moved", False)
        self.wearable: bool = kwargs.get("wearable", False)
        self.text: str = kwargs.get("text", None)
        self.reveals: str = kwargs.get("reveals", None)
        self.when_broken_do: str = kwargs.get("when_broken_do", None)
        """Value must be a method name existing in activator_handler"""
        self.enterable: bool = kwargs.get("enterable", False)
        super().__init__(name, description, **kwargs)



"""Rope module"""
from game.model.element import Element
from game.model.thing import Thing


class Rope(Thing, Element):
    """Class to initialize any kind of rope with its own name and description"""
    def __init__(self, name, description, **kwargs):
        self.name: str = name
        self.description: str = description
        self.tied_to: str = kwargs.get("tied_to", None)
        Thing.__init__(self, name, description, **kwargs)
        Element.__init__(self, name, description, **kwargs)



"""Container module"""
from game.model.element import Element
from game.model.thing import Thing


class Chest(Thing, Element):
    """Class to initialize any kind of container, e.g. a chest, with its own name and description"""
    def __init__(self, name, description, **kwargs):
        self.open: bool = kwargs.get("open", False)
        self.locked: bool = kwargs.get("locked", False)
        self.key: str = kwargs.get("key", None)
        """Value must be name of key as a string"""
        self.peekable: bool = kwargs.get("peekable", False)
        Thing.__init__(self, name, description, **kwargs)
        Element.__init__(self, name, description, **kwargs)



"""Generates the game world"""
import os
import sys
PROJECT_PATH = os.getcwd()
SOURCE_PATH = os.path.join(
    PROJECT_PATH
)
sys.path.append(SOURCE_PATH)
from game.model.activator import Activator
from game.model.door import Door
from game.model.food import Food
from game.model.location import Location
from game.model.player import Player
from game.model.rope import Rope
from game.model.thing import Thing
from game.model.tool import Tool
from game.model.enums.activator_type import ActivatorType
from game.controller.save_controller import SaveHandler
from game.data.texts import UP, DOWN, EAST, WEST
from game.data.scenario_texts import BED_DESCRIPTION, BED_NAME, BEDROOM_BRIEF, BEDROOM_BUTTON_DESCRIPTION,\
BEDROOM_BUTTON_NAME, BEDROOM_DESCRIPTION, BEDROOM_DOOR_DESCRIPTION,\
BEDROOM_DOOR_NAME, BEDROOM_HOOK_DESCRIPTION, BEDROOM_HOOK_NAME, BEDROOM_KEY_DESCRIPTION,\
BEDROOM_KEY_NAME, BEDROOM_KEY_TEXT, BEDROOM_NAME, BEDROOM_PILE_OF_DUST_DESCRIPTION,\
BEDROOM_PILE_OF_DUST_NAME, BEDROOM_RUG_DESCRIPTION, BEDROOM_RUG_NAME,\
BREAKFAST_KNIFE_DESCRIPTION, BREAKFAST_KNIFE_NAME, CELLAR_ALTAR_DESCRIPTION, CELLAR_ALTAR_NAME, CELLAR_BRIEF,\
CELLAR_DESCRIPTION, CELLAR_NAME, CELLAR_SKELETON_DESCRIPTION,\
CELLAR_SKELETON_NAME, DINING_ROOM_BRIEF, DINING_ROOM_CRATE_DESCRIPTION, DINING_ROOM_CRATE_NAME,\
DINING_ROOM_DESCRIPTION, DINING_ROOM_FIREPLACE_DESCRIPTION, DINING_ROOM_FIREPLACE_NAME,\
DINING_ROOM_FIREPLACE_SOUND, DINING_ROOM_FOOD_DESCRIPTION, DINING_ROOM_FOOD_NAME,\
DINING_ROOM_FOOD_SMELL, DINING_ROOM_FOOD_TASTE, DINING_ROOM_NAME,\
DINING_ROOM_PLATE_DESCRIPTION, DINING_ROOM_PLATE_NAME, DINING_ROOM_TABLE_DESCRIPTION,\
DINING_ROOM_TABLE_NAME, DINING_ROOM_TRAPDOOR_DESCRIPTION, DINING_ROOM_TRAPDOOR_NAME,\
PLAYER_DESCRIPTION, PLAYER_NAME, ROPE_DESCRIPTION, ROPE_NAME


all_name_locations = []
# Create all game elements
player = Player(PLAYER_NAME, PLAYER_DESCRIPTION)
player.health = 100
bedroom = Location(BEDROOM_NAME, BEDROOM_DESCRIPTION)
bedroom.brief = BEDROOM_BRIEF
bedroom.exits[WEST] = DINING_ROOM_NAME
bed = Thing(BED_NAME, BED_DESCRIPTION)
bed.fixed = True
bed.enterable = True
bedroom_door = Door(BEDROOM_DOOR_NAME, BEDROOM_DOOR_DESCRIPTION)
bedroom_door.connects.append(DINING_ROOM_NAME)
bedroom_door.locked = True
bedroom_key = Thing(BEDROOM_KEY_NAME, BEDROOM_KEY_DESCRIPTION)
bedroom_key.text = BEDROOM_KEY_TEXT
bedroom_door.key = BEDROOM_KEY_NAME
bedroom_hook = Thing(BEDROOM_HOOK_NAME, BEDROOM_HOOK_DESCRIPTION)
bedroom_hook.fixed = True
bedroom_hook.contents.append(bedroom_key)
bedroom_rug = Thing(BEDROOM_RUG_NAME, BEDROOM_RUG_DESCRIPTION)
bedroom_button = Activator(BEDROOM_BUTTON_NAME, BEDROOM_BUTTON_DESCRIPTION)
bedroom_button.type = ActivatorType.PRESS
bedroom_button.turn_on_method_name = "bedroom_button_on"
bedroom_button.turn_off_method_name = "bedroom_button_off"
bedroom_pile_of_dust = Thing(BEDROOM_PILE_OF_DUST_NAME, BEDROOM_PILE_OF_DUST_DESCRIPTION)
bedroom_pile_of_dust.visible = False
bedroom_rug.reveals = bedroom_pile_of_dust.name
bedroom_door.connects.append(BEDROOM_NAME)
bedroom.contents.extend([player, bed, bedroom_door, bedroom_hook,\
    bedroom_rug, bedroom_button, bedroom_pile_of_dust])
dining_room = Location(DINING_ROOM_NAME, DINING_ROOM_DESCRIPTION)
dining_room.brief = DINING_ROOM_BRIEF
dining_room.exits[EAST] = BEDROOM_NAME
dining_room.exits[DOWN] = CELLAR_NAME
dining_room_table = Thing(DINING_ROOM_TABLE_NAME, DINING_ROOM_TABLE_DESCRIPTION)
dining_room_table.preposition = "on"
dining_room_plate = Thing(DINING_ROOM_PLATE_NAME, DINING_ROOM_PLATE_DESCRIPTION)
dining_room_plate.preposition = "on"
dining_room_breakfast = Food(DINING_ROOM_FOOD_NAME, DINING_ROOM_FOOD_DESCRIPTION)
dining_room_breakfast.taste = DINING_ROOM_FOOD_TASTE
dining_room_breakfast.smell = DINING_ROOM_FOOD_SMELL
dining_room_plate.contents.append(dining_room_breakfast)
dining_room_trapdoor = Door(DINING_ROOM_TRAPDOOR_NAME, DINING_ROOM_TRAPDOOR_DESCRIPTION)
rope = Rope(ROPE_NAME, ROPE_DESCRIPTION)
dining_room_crate = Thing(DINING_ROOM_CRATE_NAME, DINING_ROOM_CRATE_DESCRIPTION)
dining_room_crate.fixed = True
breakfast_knife = Tool(BREAKFAST_KNIFE_NAME, BREAKFAST_KNIFE_DESCRIPTION)
dining_room_plate.when_broken_do = "break_plate"
dining_room_fireplace = Thing(DINING_ROOM_FIREPLACE_NAME, DINING_ROOM_FIREPLACE_DESCRIPTION)
dining_room_fireplace.sound = DINING_ROOM_FIREPLACE_SOUND
dining_room_table.contents.extend([dining_room_plate, breakfast_knife])
dining_room.contents.extend([dining_room_table, dining_room_trapdoor,\
    rope, dining_room_crate, dining_room_fireplace])
cellar = Location(CELLAR_NAME, CELLAR_DESCRIPTION)
cellar.brief = CELLAR_BRIEF
cellar.exits[UP] = dining_room.name
cellar.needs_rope = True
cellar_altar = Thing(CELLAR_ALTAR_NAME, CELLAR_ALTAR_DESCRIPTION)
cellar_skeleton = Thing(CELLAR_SKELETON_NAME, CELLAR_SKELETON_DESCRIPTION)
cellar.contents.extend([cellar_altar, cellar_skeleton])

# add all rooms to all_name_rooms list and save
all_name_locations.append((bedroom.name, bedroom))
all_name_locations.append((dining_room.name, dining_room))
all_name_locations.append((cellar.name, cellar))
save_handler = SaveHandler()
save_handler.save(all_name_locations, "game/data/scenario.json")



"""dungeon master module"""
import re
import pkg_resources
from typing import Tuple
from game.model.element import Element
from game.model.location import Location
from game.model.activator import Activator
from game.model.animate import Animate
from game.model.chest import Chest
from game.model.door import Door
from game.model.player import Player
from game.model.rope import Rope
from game.model.thing import Thing
from game.model.tool import Tool
from game.model.enums.activator_type import ActivatorType
from game.helper_methods import isinstanceorsubclass
from game.controller.activator_controller import ActivatorHandler
from game.controller.save_controller import SaveHandler
from game.data.texts import ACTION_FAILED, ACTION_NOT_POSSIBLE, ALREADY_OFF, ALREADY_ON,\
    ALREADY_UNTIED, APPEARING, CANT_BREAK, CANT_SEE_LOCATION_FROM_HIDING,\
    CANT_TIE_TO_ELEMENT, CLIMBING_DOWN, CLOSED, DONE, DOWN,\
    FAILED_SAVE_MESSAGE, KEY_MISSING, LOADED_SAVE_MESSAGE, LOCATION_PREFIX,\
    LOCATION_SUFFIX, NEEDS_TO_BE_TOOL, NO_SMELLS, NO_TIED_ROPE, NOT_ENTERABLE,\
    NOT_HIDING, NOT_OPENABLE, NOT_READABLE, NOTHING_HAPPENS, SAVED_GAME_MESSAGE,\
    SILENCE, SPECIFIY_HIDING_PLACE, TARGET_NOT_SPECIFIED, THAT_WONT_HOLD,\
    THREW_AT_NOTHING, TURNED_OFF, TURNED_ON, UNTIE, WEAPON_NOT_SPECIFIED,\
    door_leads_to, door_not_locked, door_unlocked, GENERIC_LOCATAION_NAME,\
    INVALID_DIRECTION, LOCKED_DOOR, eat_food, element_not_found, element_not_in_inventory,\
    entering_thing, hit_target, noises_description, picked_up_element,\
    element_in_container, reveal_element, smell_description, tie_rope_to_target


class DungeonMaster:
    """dungeon master class"""
    def __init__(self):
        self.player_location = None
        self.all_name_locations = []
        self.save_handler = SaveHandler()
        self.player_score = 0
        self.activator_handler = ActivatorHandler()
    
    def brief(self, room_name: str):
        for name_location in self.all_name_locations:
            if name_location[0] == room_name and not name_location[1].visited and name_location[1].brief:
                name_location[1].visited = True
                return name_location[1].brief + "\n\n"
        return ""

    def get_player(self):
        """Get the player object from the current location"""
        return self.get_element_container("Player", self.player_location)[0]

    def get_score(self):
        """Get the current score"""
        return self.player_score

    def get_health(self):
        """Get the current health"""
        return self.get_player().health

    def move_player(self, direction: str):
        """Move the player from one location to the next, which lies in the given direction"""
        if direction in self.player_location.exits:
            travel_text = ""
            all_name_rooms_dict =  dict(self.all_name_locations)
            next_room = all_name_rooms_dict[self.player_location.exits[direction]]
            room_has_attached_rope = False
            for element in self.player_location.contents:
                if isinstanceorsubclass(element, Door):
                    if next_room.name in element.connects and element.locked:
                        return LOCKED_DOOR
                    element.open = True
                if isinstanceorsubclass(element, Rope) and direction == DOWN and next_room.needs_rope:
                    if not element.tied_to:
                        return NO_TIED_ROPE
                    room_has_attached_rope = True
                    travel_text = CLIMBING_DOWN + "\n"
            if next_room.needs_rope and room_has_attached_rope is False:
                return NO_TIED_ROPE
            player = self.get_player()
            self.player_location.contents.remove(player)
            self.set_player_location(player, next_room)
            brief_text = ""
            brief_text = self.brief(next_room.name)
            return f"{travel_text}{brief_text}{self.player_location.name}"
        return INVALID_DIRECTION

    def unlock(self, door_name: str):
        """Unlocks and opens a door"""
        door = self.get_element_container(door_name, self.player_location)[0]
        if not door.locked:
            return door_not_locked(door_name)
        if any(element.name == door.key for element in self.get_player().contents):
            self.player_score = self.player_score + 1
            door.locked = False
            return door_unlocked(door_name)
        return KEY_MISSING

    def set_player_location(self, player: Player, location: Location):
        """Update the players location to a new one"""
        location.contents.append(player)
        self.player_location = location

    def describe(self, element_name: str):
        """Returns a desription of any kind of in-game element at the location of the player"""
        element_name = element_name.lower()
        if element_name in (self.player_location.name, GENERIC_LOCATAION_NAME, ''):
            return self.describe_location()
        return self.describe_element(element_name)

    def describe_location(self):
        """Describes the location where the player is"""
        if not self.get_player().hiding:
            return self.describe_container(self.player_location, (LOCATION_PREFIX, LOCATION_SUFFIX))
        return CANT_SEE_LOCATION_FROM_HIDING

    def describe_element(self, element_name: str):
        """Get description of a single element"""
        element_container = self.get_element_container(element_name, self.player_location)
        if element_container:
            return self.describe_container(element_container[0])
        else:
            return element_not_found(element_name)

    def describe_container(self, top_container: Element, prefix_suffix=None):
        """Descsribe any container and its contents and all the contents contents etc."""
        if prefix_suffix:
            description = top_container.name + "\n" + prefix_suffix[0] +\
                top_container.description.lower() + prefix_suffix[1]
        else:
            description = top_container.description
        visible_elements = self.get_all_elements_container(top_container, only_visible=True)
        for element_container in visible_elements:
            if element_container[1] is not self.get_player():
                if element_container[1] is not self.player_location:
                    description = description + "\n" +\
                        element_in_container(element_container[0].name,\
                        element_container[1].preposition,\
                        element_container[1].name)
                else:
                    if not isinstanceorsubclass(element_container[0], Player):
                        description = description + "\n" + element_container[0].description
                    if isinstanceorsubclass(element_container[0], Door):
                        description = description + " " + door_leads_to(self.get_door_directions(element_container))
        return description

    def get_door_directions(self, door_container: Tuple[Door, Element]):
        """Get the directions a door leads to"""
        directions = [] 
        for direction_location in door_container[1].exits.items():
            if direction_location[1] in door_container[0].connects:
                directions.append(direction_location[0])
        return directions

    def get_element_container(self, compare_element_name: str, container: Element, only_visible: bool = True):
        """Get an element in the container
        by its name and the corresponding container.
        If it's not found it returns None"""
        vague_matches = []
        for element_container in self.get_all_elements_container(container, only_visible):
            element = element_container[0]
            if bool(set(compare_element_name.lower().split()) & set(element.name.lower().split())):
                vague_matches.append((element, element_container[1]))
        if len(vague_matches) > 1:
            for vague_match in vague_matches:
                if vague_match[0].name.lower() == compare_element_name.lower():
                    return vague_match
        if len(vague_matches) == 1:
            return vague_matches[0]
        return None

    def get_all_elements_container(self, container: Element, only_visible: bool = False, only_takeable: bool = False):
        """Recursive method to get all elements in the container
        and their corresponding container. If there is nothing it returns None"""
        elements_container = []
        if only_takeable:
            for element in container.contents:
                if isinstanceorsubclass(element, Thing) and element.visible and not element.fixed:
                    elements_container.append((element, container))
                elements_container = elements_container +\
                    self.get_all_elements_container(element, only_visible, only_takeable)
            return elements_container
        if only_visible:
            for element in container.contents:
                if element.visible:
                    elements_container.append((element, container))
                if isinstanceorsubclass(element, Chest) and not element.peekable or\
                isinstanceorsubclass(element, Chest) and not element.open:
                    pass
                else:
                    elements_container = elements_container +\
                        self.get_all_elements_container(element, only_visible)
            return elements_container
        if not only_visible:
            for element in container.contents:
                elements_container.append((element, container))
                elements_container = elements_container +\
                    self.get_all_elements_container(element, only_visible)
            return elements_container
        return None

    def take(self, element_name: str):
        """Take all elements or a single element and put it into the players inventory"""
        if element_name == "all":
            all_takeable = self.get_all_elements_container(self.player_location, only_takeable=True)
            response = ""
            for element_container in all_takeable:
                if not isinstanceorsubclass(element_container[0], Player) and not element_container[0].fixed:
                    element_container[1].contents.remove(element_container[0])
                    self.get_player().contents.append(element_container[0])
                    response = response + picked_up_element(element_container[0].name) + "\n"
            return response
        element_container = self.get_element_container(element_name, self.player_location)
        if not element_container:
            return element_not_found(element_name)
        element_container[1].contents.remove(element_container[0])
        self.get_player().contents.append(element_container[0])
        return picked_up_element(element_container[0].name)

    def get_player_inventory(self):
        """Returns the invenotry of the player as a string listing all things"""
        description = ""
        for element in self.get_player().contents:
            description = description + element.name + "\n"
        return description

    def save(self):
        """Saves the game state to file"""
        if self.save_handler.save(self.all_name_locations, pkg_resources.resource_filename("game.data", "save.json")):
            return SAVED_GAME_MESSAGE
        return FAILED_SAVE_MESSAGE

    def load(self, filename: str):
        """Loads the game state from file"""
        load_data = self.save_handler.load(filename)
        self.all_name_locations = load_data[0]
        self.player_location = load_data[1]
        return LOADED_SAVE_MESSAGE

    def throw(self, instructions: str):
        """Throws an item"""
        if "at" in instructions:
            thing_target = instructions.split(" at ")
            thing_container = self.get_element_container(thing_target[0], self.get_player())[0]
            if thing_container is None:
                return element_not_found(thing_container.name)
            target_container = self.get_element_container(thing_target[1], self.player_location)[0]
            if target_container is None:
                return element_not_found(target_container.name)
            if isinstanceorsubclass(target_container, Animate):
                self.get_player().contents.remove(thing_container)
                target_container.health = target_container.health - thing_container.damage * 1.5
                self.player_location.contents.append(thing_container)
                return hit_target(target_container.name)
        thing_container = self.get_element_container(instructions, self.player_location)
        self.get_player().contents.remove(thing_container[0])
        self.player_location.contents.append(thing_container[0])
        return THREW_AT_NOTHING

    def close(self, element_name: str):
        """Closes a door or chest element"""
        element_container = self.get_element_container(element_name, self.player_location)
        if isinstanceorsubclass(element_container[0], (Chest, Door)):
            element_container[0].open = False
            return CLOSED
        return NOT_OPENABLE

    def read(self, element_name: str):
        """Read an elements text"""
        element_container = self.get_element_container(element_name, self.player_location)
        if isinstanceorsubclass(element_container[0], Thing) and element_container[0].text:
            return element_container[0].text
        return NOT_READABLE

    def put(self, instructions: str):
        """Put element into the contents of another element"""
        if " in " in instructions or " on " in instructions:
            thing_target = re.split(" in | on ", instructions)
            thing_container = self.get_element_container(thing_target[0], self.get_player())
            if thing_container is None:
                return element_not_found(thing_target[0])
            target_container = self.get_element_container(thing_target[1], self.player_location)
            if target_container is None:
                return element_not_found(thing_target[1])
            thing_container[1].contents.remove(thing_container[0])
            target_container[0].contents.append(thing_container[0])
            return DONE
        return TARGET_NOT_SPECIFIED

    def activate(self, instructions: str, expected_activator_type: ActivatorType):
        """Activate an activator"""
        split_input = instructions.split()
        if "on" in instructions.split():
            split_input.remove("on")
            activator = self.get_element_container(split_input[0], self.player_location)[0]
            if not self.turn_on(activator, self.activator_handler):
                return ALREADY_ON
            return TURNED_ON
        if "off" in instructions.split():
            split_input.remove("off")
            activator = self.get_element_container(split_input[0], self.player_location)[0]
            if not self.turn_off(activator, self.activator_handler):
                return ALREADY_OFF
            return TURNED_OFF
        activator = self.get_element_container(instructions, self.player_location)[0]
        if not activator:
            return element_not_found(instructions)
        if not expected_activator_type is activator.type:
            return ACTION_NOT_POSSIBLE
        if activator.is_on is True:
            return self.turn_off(activator, self.activator_handler)
        if activator.is_on is False:
            return self.turn_on(activator, self.activator_handler)
        return ACTION_FAILED

    def turn_on(self, activator: Activator, activator_handler: ActivatorHandler):
        """Turns an activator on"""
        if activator.is_on:
            return ALREADY_ON
        turn_on_method = getattr(activator_handler, activator.turn_on_method_name)
        activator.is_on = True
        if turn_on_method:
            return turn_on_method()
        return NOTHING_HAPPENS

    def turn_off(self, activator: Activator, activator_handler: ActivatorHandler):
        """Turns an activator off"""
        if not activator.is_on:
            return ALREADY_OFF
        turn_off_method = getattr(activator_handler, activator.turn_off_method_name)
        activator.is_on = False
        if turn_off_method:
            return turn_off_method()
        return NOTHING_HAPPENS

    def move_element(self, element_name: str):
        """Moves element"""
        element_container = self.get_element_container(element_name, self.player_location)
        if not element_container:
            return element_not_found(element_name)
        element_container[0].moved = True
        revealed_element = self.get_element_container(element_container[0].reveals,\
            self.player_location, only_visible=False)[0]
        revealed_element.visible = True
        return reveal_element(element_name, revealed_element.description.lower())

    def attack(self, instructions: str):
        """Attacks a target"""
        if "with" in instructions:
            target_thing = instructions.split(" with ")
            tool_container = self.get_element_container(target_thing[1], self.get_player())
            if tool_container is None:
                return element_not_in_inventory(target_thing[1])
            if not isinstanceorsubclass(tool_container[0], Tool):
                return NEEDS_TO_BE_TOOL
            target_container = self.get_element_container(target_thing[0], self.player_location)
            if target_container is None:
                return element_not_found(target_thing[0])
            if isinstanceorsubclass(target_container[0], Animate):
                target_container[0].health = target_container[0].health - tool_container[0].damage
                return hit_target(target_container[0].name)
            if isinstanceorsubclass(target_container[0], Thing):
                if not target_container[0].when_broken_do:
                    return CANT_BREAK
                break_method = getattr(self.activator_handler, target_container[0].when_broken_do)
                target_container[1].contents.remove(target_container[0])
                target_container[1].contents.append(\
                    Thing("broken " + target_container[0].name,\
                    target_container[0].description))
                return break_method()
        return WEAPON_NOT_SPECIFIED

    def eat(self, element_name: str):
        """Feeds the player"""
        food = self.get_element_container(element_name, self.player_location)[0]
        if not food:
            return element_not_found(food)
        self.get_player().health += food.regen
        return eat_food(food.name, food.taste)

    def tie(self, instructions: str):
        """Ties a rope to something"""
        if "to" in instructions:
            thing_target = instructions.split(" to ")
            thing_container = self.get_element_container(thing_target[0], self.get_player())
            if thing_container is None:
                return element_not_in_inventory(thing_target[0])
            target_container = self.get_element_container(thing_target[1], self.player_location)
            if target_container is None:
                return element_not_found(thing_target[1])
            if not isinstanceorsubclass(target_container[0], Thing):
                return CANT_TIE_TO_ELEMENT
            if not target_container[0].fixed:
                return THAT_WONT_HOLD
            if isinstanceorsubclass(thing_container[0], Rope):
                self.get_player().contents.remove(thing_container[0])
                target_container[1].contents.append(thing_container[0])
                thing_container[0].tied_to = target_container[0].name
                return tie_rope_to_target(target_container[0].name.lower())
        return TARGET_NOT_SPECIFIED

    def untie(self, rope_name: str):
        """Unties a rope"""
        rope = self.get_element_container(rope_name, self.player_location)
        if rope is None:
            return element_not_found(rope_name)
        if self.get_element_container(rope.tied_to, self.player_location):
            rope.tied_to = None
            return UNTIE
        return ALREADY_UNTIED

    def listen(self):
        """Listens to environment"""
        noises = []
        for element_container in self.get_all_elements_container(self.player_location):
            if element_container[0].sound:
                noises.append(element_container[0].sound)
        if len(noises) == 0:
            return SILENCE
        return noises_description(noises)

    def smell(self):
        """Smells the environment"""
        smells = []
        for element_container in self.get_all_elements_container(self.player_location):
            if element_container[0].smell:
                smells.append(element_container[0].smell)
        if len(smells) == 0:
            return NO_SMELLS
        return smell_description(smells)

    def hide(self, instructions: str):
        """Hides player in thing"""
        if "in" in instructions or "under" in instructions:
            target = re.split("in|under", instructions)[1]
            element_container = self.get_element_container(target, self.player_location)
            if isinstanceorsubclass(element_container[0], Thing) and element_container[0].enterable:
                self.get_player().hiding = True
                return entering_thing(element_container[0].name)
            return NOT_ENTERABLE
        return SPECIFIY_HIDING_PLACE

    def appear(self):
        """Puts player out of hiding"""
        player = self.get_player()
        if player.hiding is True:
            player.hiding = False
            return APPEARING
        return NOT_HIDING



"""Generate test world module"""
import os
import sys
PROJECT_PATH = os.getcwd()
SOURCE_PATH = os.path.join(
    PROJECT_PATH
)
sys.path.append(SOURCE_PATH)
import json
from game.model.animate import Animate
from game.model.chest import Chest
from game.model.element import Element
from game.model.activator import Activator
from game.model.door import Door
from game.model.food import Food
from game.model.location import Location
from game.model.player import Player
from game.model.rope import Rope
from game.model.thing import Thing
from game.model.tool import Tool
from game.model.enums.activator_type import ActivatorType
from game.helper_methods import isinstanceorsubclass

test_elements = [
    Activator("test", "test"),
    Animate("test", "test"),
    Chest("test", "test"),
    Door("test", "test"),
    Food("test", "test"),
    Element("test", "test"),
    Player("test", "test"),
    Rope("test", "test"),
    Thing("test", "test"),
    Tool("test", "test"),
    Location("test", "test")
]

for element in test_elements:
    if isinstanceorsubclass(element, Activator):
        element.type = ActivatorType.PRESS
        element.is_on = True
        element.turn_on_method_name = "test_turn_on_method"
        element.turn_off_method_name = "test_turn_off_method"
    if isinstanceorsubclass(element, Animate):
        element.clothes = ["test"]
        element.health = 50
    if isinstanceorsubclass(element, Chest):
        element.open = True
        element.locked = True
        element.key = "test"
        element.peekable = True
    if isinstanceorsubclass(element, Door):
        element.open = True
        element.lockable = True
        element.locked = True
        element.key = "test"
        element.connects = ["test"]
    if isinstanceorsubclass(element, Food):
        element.regen = 50
        element.taste = "test"
    if isinstanceorsubclass(element, Player):
        element.hiding = True
    if isinstanceorsubclass(element, Rope):
        element.tied_to = "test"
    if isinstanceorsubclass(element, Thing):
        element.when_broken_do = "test_broken_method"
        element.reveals = "test"
        element.moved = True
        element.fixed = True
        element.visible = True
        element.enterable = True
        element.wearable = True
        element.text = "test"
    if isinstanceorsubclass(element, Tool):
        element.damage = 50
        element.durability = 50
    if isinstanceorsubclass(element, Element):
        element.visible = True
        element.preposition = "test"
        element.sound = "test"
        element.smell = "test"
        element.contents = [Element("test", "test")]

test_location = Location("test", "test")
test_location.exits = {"test": "test"}
test_location.brief = "test"
test_location.has_light = True
test_location.needs_rope = True
test_location.visited = True
test_location.contents.extend(test_elements)

class ElementEncoder(json.JSONEncoder):
    """json encoder for elements"""
    def default(self, o):
        return o.__dict__

with open("tests/test_world.json", "w", encoding="UTF-8") as savefile:
    json.dump([test_location], savefile, indent=4, cls=ElementEncoder)



"""Player module"""
from game.model.animate import Animate
from game.model.element import Element


class Player(Animate, Element):
    """Class to initialize the player with their own name and description"""
    def __init__(self, name, description, **kwargs):
        self.name: str = name
        self.description: str = description
        self.visible: bool = kwargs.get("visible", True)
        self.hiding: bool = kwargs.get("hiding", False)
        Animate.__init__(self, name, description, **kwargs)
        Element.__init__(self, name, description, **kwargs)



"""activator handler module"""
from game.data.scenario_texts import BUTTON_GOES_DARK, BUTTON_LIGHTS_UP, PLATE_BREAKS


class ActivatorHandler():
    """Class containing the methods corresponding to button presses,
    valve turns or thing breaking"""
    def __init__(self):
        pass

    def bedroom_button_on(self):
        """Turns on bedroom test button"""
        return BUTTON_LIGHTS_UP

    def bedroom_button_off(self):
        """Turns off bedroom test button"""
        return BUTTON_GOES_DARK

    def break_plate(self):
        """Reacts to the player breaking a plate"""
        return PLATE_BREAKS



"""console view module"""
from functools import partial
import random
import pkg_resources
from game.model.enums.activator_type import ActivatorType
from game.data.scenario_texts import BEDROOM_NAME
from game.data.texts import DOWN, EAST, GREETINGS, INPUT_INDICATOR, INTRODUCTION, JUMP_RESPONSE,\
    NORTH, NORTHEAST, NORTHWEST, NOTHING_RESPONSES, PLEASE_TYPE,\
    QUIT_MESSAGE, SHOUT_RESPONSE, SOUTH, SOUTHEAST,\
    SOUTHWEST, SWEAR_RESPONSE, UP, WEST


class ConsoleView:
    """Class to initialize a view for the console"""

    def __init__(self, dungeon_master):
        self.quit = False
        self.dungeon_master = dungeon_master

    def toggle_quit(self):
        """Quits the game"""
        self.quit = True
        return QUIT_MESSAGE

    def parse(self, user_input: str):
        """Runtime polymorphic method to map the possible user input to a corresponding method"""
        split_user_input = user_input.lower().split()
        if len(split_user_input) < 2:
            split_user_input.append("")
        rest_input_joined = " ".join(split_user_input[1:])
        describe = self.dungeon_master.describe
        move = self.dungeon_master.move_player
        unlock = self.dungeon_master.unlock
        take = self.dungeon_master.take
        inventory = self.dungeon_master.get_player_inventory
        save = self.dungeon_master.save
        load = self.dungeon_master.load
        score = self.dungeon_master.get_score
        health = self.dungeon_master.get_health
        throw = self.dungeon_master.throw
        close = self.dungeon_master.close
        read = self.dungeon_master.read
        put = self.dungeon_master.put
        activate = self.dungeon_master.activate
        move_element = self.dungeon_master.move_element
        attack = self.dungeon_master.attack
        eat = self.dungeon_master.eat
        tie = self.dungeon_master.tie
        untie = self.dungeon_master.untie
        listen = self.dungeon_master.listen
        smell = self.dungeon_master.smell
        hide = self.dungeon_master.hide
        appear = self.dungeon_master.appear
        move_dictionary = {
            "examine": partial(describe, rest_input_joined),
            "look": partial(describe, rest_input_joined),
            "l": partial(describe, rest_input_joined),
            "north": partial(move, NORTH),
            "n": partial(move, NORTH),
            "east": partial(move, EAST),
            "e": partial(move, EAST),
            "south": partial(move, SOUTH),
            "s": partial(move, SOUTH),
            "west": partial(move, WEST),
            "w": partial(move, WEST),
            "northeast": partial(move, NORTHEAST),
            "ne": partial(move, NORTHEAST),
            "northwest": partial(move, NORTHWEST),
            "nw": partial(move, NORTHWEST),
            "southeast": partial(move, SOUTHEAST),
            "se": partial(move, SOUTHEAST),
            "southwest": partial(move, SOUTHWEST),
            "sw": partial(move, SOUTHWEST),
            "up": partial(move, UP),
            "u": partial(move, UP),
            "climb": partial(move, UP),
            "down": partial(move, DOWN),
            "d": partial(move, DOWN),
        }
        general_dictionary = {
            "quit": self.toggle_quit,
            "q": self.toggle_quit,
            "exit": self.toggle_quit,
            "go": move_dictionary.get(rest_input_joined, None),
            "open": partial(unlock, rest_input_joined),
            "unlock": partial(unlock, rest_input_joined),
            "get": partial(take, rest_input_joined),
            "take": partial(take, rest_input_joined),
            "pick": partial(take, " ".join(split_user_input[2:])),
            "inventory": inventory,
            "i": inventory,
            "hi": partial(random.choice, GREETINGS),
            "hello": partial(random.choice, GREETINGS),
            "shit": SWEAR_RESPONSE,
            "damn": SWEAR_RESPONSE,
            "fuck": SWEAR_RESPONSE,
            "jump": JUMP_RESPONSE,
            "save": save,
            "load": partial(load, pkg_resources.resource_filename("game.data", "save.json")),
            "restore": partial(load, pkg_resources.resource_filename("game.data", "save.json")),
            "restart": partial(load, pkg_resources.resource_filename("game.data", "scenario.json")),
            "score": score,
            "diagnostic": health,
            "health": health,
            "": partial(random.choice, NOTHING_RESPONSES),
            "hemonaluto": INTRODUCTION,
            "throw": partial(throw, rest_input_joined),
            "close": partial(close, rest_input_joined),
            "read": partial(read, rest_input_joined),
            "drop": partial(throw, rest_input_joined),
            "put": partial(put, rest_input_joined),
            "press": partial(activate, rest_input_joined, ActivatorType.PRESS),
            "turn": partial(activate, rest_input_joined, ActivatorType.TURN),
            "push": partial(move_element, rest_input_joined),
            "pull": partial(move_element, rest_input_joined),
            "move": partial(move_element, rest_input_joined),
            "attack": partial(attack, rest_input_joined),
            "kill": partial(attack, rest_input_joined),
            "eat": partial(eat, rest_input_joined),
            "drink": partial(eat, rest_input_joined),
            "shout": SHOUT_RESPONSE,
            "tie": partial(tie, rest_input_joined),
            "attach": partial(tie, rest_input_joined),
            "untie": partial(untie, rest_input_joined),
            "destroy": partial(attack, rest_input_joined),
            "break": partial(attack, rest_input_joined),
            "cut": partial(attack, rest_input_joined),
            "listen": listen,
            "smell": smell,
            "hide": partial(hide, rest_input_joined),
            "leave": appear,
            "appear": appear
        }
        move_action = move_dictionary.get(split_user_input[0], None)
        if move_action is not None:
            return move_action()
        general_action = general_dictionary.get(split_user_input[0], None)
        if general_action is not None:
            return general_action()

    def start_view(self):
        """Start the process of displaying messages to the cosole"""
        with open(pkg_resources.resource_filename("game.data", "logo.txt"), "r", encoding="UTF-8") as logo_file:
            print(logo_file.read())
        print(INTRODUCTION + "\n")
        print(self.dungeon_master.brief(BEDROOM_NAME))
        print(PLEASE_TYPE)
        while self.quit is False:
            user_input = input(INPUT_INDICATOR)
            print(self.parse(user_input))
            


"""Food module"""
from game.model.element import Element
from game.model.thing import Thing


class Food(Thing, Element):
    """Class to initialize any kind of Food, e.g. a baguette, with its own name and description"""
    def __init__(self, name, description, **kwargs):
        self.name: str = name
        self.description: str = description
        self.regen: int = kwargs.get("regen", 10)
        self.taste: str = kwargs.get("taste", None)
        Thing.__init__(self, name, description, **kwargs)
        Element.__init__(self, name, description, **kwargs)



import pathlib
import os
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

extra_files = package_files('game/data')

setup(
    install_requires=["parameterized==0.8.1"],
    name = "hemonaluto",
    version = "0.0.1",
    author = "Gabriel Schafflützel",
    author_email = "schaffluetzel.gabriel@gmail.com",
    description = ("A text adventure game!"),
    license = "GNU General Public License v3.0",
    keywords = "text-based adventure game",
    url = "https://github.com/gasupidupi/Hemonaluto",
    long_description=README,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    package_data={"": extra_files},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Topic :: Games/Entertainment",
        "Environment :: Console",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    python_requires=">=3.8",
    entry_points={"console_scripts": ["hemonaluto=game.runner:run", ], },
)



"""test save handler module"""
import unittest
from game.model.activator import Activator
from game.model.animate import Animate
from game.model.chest import Chest
from game.model.door import Door
from game.model.element import Element
from game.model.food import Food
from game.model.location import Location
from game.model.player import Player
from game.model.rope import Rope
from game.model.thing import Thing
from game.model.tool import Tool
from game.model.enums.activator_type import ActivatorType
from game.controller.save_controller import SaveHandler


class TestSaveHandler(unittest.TestCase):
    """Test SaveHandler class"""

    def setUp(self):
        """Set up environment required for tests"""
        self.save_handler = SaveHandler()
        self.names_locations, self.player_location = self.save_handler.load("tests/test_world.json")

    def test_load_names_locations(self):
        """Test if it loads names_locations correctly"""
        name = self.names_locations[0][0]
        location: Location = self.names_locations[0][1]
        self.assertEqual("test", name)
        self.assertEqual("test", location.name)
        self.assertEqual("test", location.description)
        self.assertEqual("Location", location.class_name)
        self.assertEqual("test", location.brief)
        self.assertEqual(True, location.has_light)
        self.assertEqual(True, location.needs_rope)
        self.assertEqual(11, len(location.contents))
        self.assertEqual("test", location.exits["test"])

    def test_load_contents(self):
        """Test if all elements get loaded correctly"""
        location: Location = self.names_locations[0][1]
        for element in location.contents:
            if isinstance(element, Activator) or issubclass(element.__class__, Activator):
                self.assertEqual(element.type, ActivatorType.PRESS)
                self.assertEqual(element.is_on, True)
                self.assertEqual(element.turn_on_method_name, "test_turn_on_method")
                self.assertEqual(element.turn_off_method_name, "test_turn_off_method")
            if isinstance(element, Animate) or issubclass(element.__class__, Animate):
                self.assertEqual(element.clothes[0], "test")
                self.assertEqual(element.health, 50)
            if isinstance(element, Chest) or issubclass(element.__class__, Chest):
                self.assertEqual(element.open, True)
                self.assertEqual(element.locked, True)
                self.assertEqual(element.key, "test")
                self.assertEqual(element.peekable, True)
            if isinstance(element, Door) or issubclass(element.__class__, Door):
                self.assertEqual(element.open, True)
                self.assertEqual(element.lockable, True)
                self.assertEqual(element.locked, True)
                self.assertEqual(element.key, "test")
                self.assertEqual(element.connects[0], "test")
            if isinstance(element, Element) or issubclass(element.__class__, Element):
                self.assertEqual(element.visible, True)
                self.assertEqual(element.contents[0].name, "test")
                self.assertEqual(element.preposition, "test")
                self.assertEqual(element.sound, "test")
                self.assertEqual(element.smell, "test")
            if isinstance(element, Food) or issubclass(element.__class__, Food):
                self.assertEqual(element.regen, 50)
                self.assertEqual(element.taste, "test")
            if isinstance(element, Player) or issubclass(element.__class__, Player):
                self.assertEqual(element.hiding, True)
            if isinstance(element, Rope) or issubclass(element.__class__, Rope):
                self.assertEqual(element.tied_to, "test")
            if isinstance(element, Thing) or issubclass(element.__class__, Thing):
                self.assertEqual(element.when_broken_do, "test_broken_method")
                self.assertEqual(element.reveals, "test")
                self.assertEqual(element.moved, True)
                self.assertEqual(element.fixed, True)
                self.assertEqual(element.visible, True)
                self.assertEqual(element.enterable, True)
                self.assertEqual(element.wearable, True)
                self.assertEqual(element.text, "test")
            if isinstance(element, Tool) or issubclass(element.__class__, Tool):
                self.assertEqual(element.damage, 50)
                self.assertEqual(element.durability, 50)



"""Animate module"""
from game.model.element import Element


class Animate(Element):
    """Class to initialize any animate being, e.g. a dwarf, with its own name and description"""
    def __init__(self, name, description, **kwargs):
        self.name: str = name
        self.description: str = description
        self.clothes: list = kwargs.get("clothes", [])
        """Values in list must be names of clothes as strings"""
        self.health: int = kwargs.get("health", 10)
        Element.__init__(self, name, description, **kwargs)
